import os
import sys
import traceback
from optparse import OptionParser

# import contracts
from zuper_commons.fs import friendly_path

from .scripts_utils import wrap_script_entry_point
from .. import CompmakeConstants, set_compmake_status, version
from ..config import config_populate_optparser
from ..context import Context
from ..exceptions import CommandFailed, CompmakeBug, MakeFailed, UserError
from ..jobs import all_jobs
from ..storage import StorageFilesystem
from ..ui import compmake_console_gui, interpret_commands_wrap
from ..ui.visualization import ui_info
from ..utils import setproctitle


# TODO: revise all of this
def read_rc_files(context: Context):
    assert context is not None
    possible = [
        "~/.compmake/compmake.rc",
        "~/.config/compmake.rc",
        "~/.compmake.rc",
        "~/compmake.rc",
        ".compmake.rc",
        "compmake.rc",
    ]
    done = False
    for x in possible:
        x = os.path.expanduser(x)
        if os.path.exists(x):
            read_commands_from_file(filename=x, context=context)
            done = True
    if not done:
        # logger.info('No configuration found (looked for %s)'
        # % "; ".join(possible))
        pass


def read_commands_from_file(filename: str, context: Context):
    from compmake.jobs.uptodate import CacheQueryDB

    filename = os.path.realpath(filename)
    if filename in context.rc_files_read:
        return
    else:
        context.rc_files_read.append(filename)

    cq = CacheQueryDB(context.get_compmake_db())

    ui_info(context, f"Reading configuration from {friendly_path(filename)}.")
    with open(filename, "r") as f:
        for line in f:
            line = line.strip()
            if not line:
                continue
            if line[0] == "#":
                continue
            interpret_commands_wrap(line, context=context, cq=cq)


usage = """
The "compmake" script takes a DB directory as argument:

    $ compmake  <compmake_storage>  [-c COMMAND]

For example:

    $ compmake out-compmake -c "clean; parmake n=2"

"""


def main():
    wrap_script_entry_point(compmake_main, exceptions_no_traceback=(UserError,))


# noinspection PyUnresolvedReferences
def compmake_main(args):
    if not "" in sys.path:
        sys.path.append("")

    setproctitle("compmake")

    parser = OptionParser(version=version, usage=usage)

    parser.add_option("--profile", default=False, action="store_true", help="Use Python profiler")

    parser.add_option("--contracts", default=False, action="store_true", help="Activate PyContracts")
    parser.add_option("--gui", default=False, action="store_true", help="Use text gui")

    parser.add_option("-c", "--command", default=None, help="Run the given command")

    parser.add_option("-n", "--namespace", default="default")

    parser.add_option(
        "--retcodefile",
        help="If given, the return value is written in this "
        "file. Useful to check when compmake finished in "
        "a grid environment. ",
        default=None,
    )

    parser.add_option(
        "--nosysexit",
        default=False,
        action="store_true",
        help="Does not sys.exit(ret); useful for debugging.",
    )

    config_populate_optparser(parser)

    (options, args) = parser.parse_args(args)

    # if not options.contracts:
    #     # info('Disabling PyContracts; use --contracts to activate.')
    #     contracts.disable_all()

    # We load plugins after we parsed the configuration
    from compmake import plugins  # @UnusedImport

    # XXX make sure this is the default
    if not args:
        msg = "I expect at least one argument (db path)." ' Use "compmake -h" for usage information.'
        raise UserError(msg)

    if len(args) >= 2:
        msg = 'I only expect one argument. Use "compmake -h" for usage ' "information."
        msg += "\n args: %s" % args
        raise UserError(msg)

    # if the argument looks like a dirname
    one_arg = args[0]
    if os.path.exists(one_arg) and os.path.isdir(one_arg):
        # If there is a compmake/ folder inside, take it as the root
        child = os.path.join(one_arg, "compmake")
        if os.path.exists(child):
            one_arg = child

        context = load_existing_db(one_arg)
        # If the context was custom we load it
        if "context" in context.compmake_db:
            context = context.compmake_db["context"]

            # TODO: check number of jobs is nonzero
    else:
        msg = "Directory not found: %s" % one_arg
        raise UserError(msg)

    args = args[1:]

    def go(context2):
        assert context2 is not None

        if options.command:
            set_compmake_status(CompmakeConstants.compmake_status_slave)
        else:
            set_compmake_status(CompmakeConstants.compmake_status_interactive)

        read_rc_files(context2)

        try:
            if options.command:
                context2.batch_command(options.command)
            else:
                if options.gui:
                    compmake_console_gui(context2)
                else:
                    context2.compmake_console()
        except MakeFailed:
            retcode = CompmakeConstants.RET_CODE_JOB_FAILED
        except CommandFailed:
            retcode = CompmakeConstants.RET_CODE_COMMAND_FAILED
        except CompmakeBug as e:
            sys.stderr.write("unexpected exception: %s" % traceback.format_exc())
            retcode = CompmakeConstants.RET_CODE_COMPMAKE_BUG
        except BaseException as e:
            sys.stderr.write("unexpected exception: %s" % traceback.format_exc())
            retcode = CompmakeConstants.RET_CODE_COMPMAKE_BUG
        except:
            retcode = CompmakeConstants.RET_CODE_COMPMAKE_BUG
        else:
            retcode = 0

        if options.retcodefile is not None:
            write_atomic(options.retcodefile, str(retcode))

        if options.nosysexit:
            return retcode
        else:
            sys.exit(retcode)

    if not options.profile:
        return go(context2=context)
    else:
        # XXX: change variables
        import cProfile

        cProfile.runctx("go(context)", globals(), locals(), "out/compmake.profile")
        import pstats

        p = pstats.Stats("out/compmake.profile")
        n = 50
        p.sort_stats("cumulative").print_stats(n)
        p.sort_stats("time").print_stats(n)


def write_atomic(filename, contents):
    dirname = os.path.dirname(filename)
    if dirname:
        if not os.path.exists(dirname):
            try:
                os.makedirs(dirname)
            except:
                pass
    tmpfile = filename + ".tmp"
    f = open(tmpfile, "w")
    f.write(contents)
    f.flush()
    os.fsync(f.fileno())
    f.close()
    os.rename(tmpfile, filename)


def load_existing_db(dirname) -> Context:
    assert os.path.isdir(dirname)
    print("Loading existing jobs DB %r." % dirname)
    # check if it is compressed
    files = os.listdir(dirname)
    for one in files:
        if ".gz" in one:
            compress = True
            break
    else:
        compress = False

    db = StorageFilesystem(dirname, compress=compress)
    context = Context(db=db)
    jobs = list(all_jobs(db=db))
    # logger.info('Found %d existing jobs.' % len(jobs))
    context.reset_jobs_defined_in_this_session(jobs)

    return context



from system_cmd import system_cmd_result
import os

__all__ = [
    "JobNotRunYet",
    "get_qacct",
]


class JobNotRunYet(Exception):
    pass


def get_qacct(sge_id):
    """
        Only provides results if the job already rann.

        Raises JobNotRunYet if the job didn't run.
    """
    cmd = ["qacct", "-j", sge_id]
    cwd = os.getcwd()
    res = system_cmd_result(
        cwd,
        cmd,
        display_stdout=False,
        display_stderr=False,
        raise_on_error=False,
        capture_keyboard_interrupt=False,
    )
    if res.ret == 1:
        if "startup" in res.stderr or "startup" in res.stdout:
            # /opt/sge6/default/common/accounting: No such file or directory
            # no jobs running since startup
            raise JobNotRunYet(sge_id)
        if "not found" in res.stderr:
            # todo
            raise JobNotRunYet(sge_id)
        raise Exception("qcct failed: %s" % res)
    if res.ret != 0:
        raise Exception("qcct failed: %s" % res)
    values = {}
    for line in res.stdout.split("\n"):
        tokens = line.split()
        if len(tokens) >= 2:  # XXX
            k = tokens[0]
            v = " ".join(tokens[1:])
            if k == "failed":
                v = tokens[1]
            values[k] = v
    return values



from nose.tools import istest

from .. import set_compmake_status, CompmakeConstants
from .compmake_test import CompmakeTest


def bottom():
    TestOrder.order.append("bottom")


def bottom2():
    TestOrder.order.append("bottom2")


def top(x):
    TestOrder.order.append("top")


@istest
class TestOrder(CompmakeTest):

    order = []

    def mySetUp(self):
        # TODO: use tmp dir
        set_compmake_status(CompmakeConstants.compmake_status_embedded)
        # clear the variable holding the result
        TestOrder.order = []

    def test_order(self):
        # add two copies
        self.comp(top, self.comp(bottom))
        self.comp(top, self.comp(bottom))

        self.cc.batch_command("clean")
        self.cc.batch_command("make")

        self.assertEqual(["bottom", "top", "bottom", "top"], TestOrder.order)

    def test_order_2(self):
        # choose wisely here
        self.comp(top, self.comp(bottom))
        self.comp(top, self.comp(bottom))
        self.comp(bottom2)

        self.cc.batch_command("clean")
        self.cc.batch_command("make")

        self.assertEqual(["bottom2", "bottom", "top", "bottom", "top"], TestOrder.order)

    def test_order_3(self):
        # choose wisely here
        self.comp(top, self.comp(bottom2))
        self.comp(bottom)
        self.comp(top, self.comp(bottom2))

        self.cc.batch_command("clean")
        self.cc.batch_command("make")

        self.assertEqual(["bottom", "bottom2", "top", "bottom2", "top"], TestOrder.order)



from typing import List

from zuper_commons.text import indent
from zuper_commons.types import raise_wrapped, ZException


class ShellExitRequested(ZException):
    pass


class CompmakeException(ZException):
    pass


class CompmakeDBError(CompmakeException):
    """ Raised for inconsistencies in the DB. """


class CompmakeBug(CompmakeException):
    def get_result_dict(self):
        res = dict(bug=str(self))
        return res

    @staticmethod
    def from_dict(res):
        from compmake.jobs.result_dict import result_dict_check

        result_dict_check(res)
        assert "bug" in res
        e = CompmakeBug(res["bug"])
        return e


class CommandFailed(ZException):
    pass


class MakeFailed(CommandFailed):
    def __init__(self, failed: List[str], blocked: List[str] = None):
        failed = failed or []
        blocked = blocked or []
        self.failed = set(failed)
        self.blocked = set(blocked)
        msg = f"Make failed ({len(self.failed)} failed, {len(self.blocked)} blocked)"
        CommandFailed.__init__(self, msg, failed=failed, blocked=blocked)


class MakeHostFailed(CommandFailed):
    # Thrown when all workers have aborted
    pass


class KeyNotFound(CompmakeException):
    pass


class UserError(CompmakeException):
    pass


class SerializationError(UserError):
    """ Something cannot be serialized (function or function result)."""

    pass


class CompmakeSyntaxError(UserError):
    pass


class JobFailed(CompmakeException):
    """ This signals that some job has failed """

    def __init__(self, job_id: str, reason: str, bt: str, deleted_jobs: List[str] = None):
        deleted_jobs = deleted_jobs or []
        self.job_id = job_id
        self.reason = reason
        self.bt = bt
        self.deleted_jobs = set(deleted_jobs)

        CompmakeException.__init__(self, job_id=job_id, reason=reason, bt=bt, deleted_jobs=deleted_jobs)

    def get_result_dict(self):
        res = dict(
            fail=f"Job {self.job_id!r} failed.",
            job_id=self.job_id,
            reason=self.reason,
            deleted_jobs=self.deleted_jobs,
            bt=self.bt,
        )
        return res

    @staticmethod
    def from_dict(res):
        from compmake.jobs.result_dict import result_dict_check

        result_dict_check(res)
        assert "fail" in res
        e = JobFailed(
            job_id=res["job_id"], bt=res["bt"], reason=res["reason"], deleted_jobs=res["deleted_jobs"]
        )
        return e


class JobInterrupted(CompmakeException):
    """ User requested to interrupt job"""

    def __init__(self, job_id, deleted_jobs: List[str] = None):
        deleted_jobs = deleted_jobs or []
        self.job_id = job_id
        self.deleted_jobs = set(deleted_jobs)

        self.msg = f"Job {self.job_id!r} received KeyboardInterrupt."
        CompmakeException.__init__(self, self.msg, job_id=job_id, deleted_jobs=deleted_jobs)

    def __str__(self):
        return self.msg

    @staticmethod
    def from_dict(res):
        from compmake.jobs.result_dict import result_dict_check

        result_dict_check(res)
        assert "interrupted" in res
        e = JobInterrupted(job_id=res["job_id"], deleted_jobs=res["deleted_jobs"])
        return e

    def get_result_dict(self):
        res = dict(
            interrupt=f"Job {self.job_id!r} interrupted.",
            job_id=self.job_id,
            deleted_jobs=sorted(self.deleted_jobs),
        )
        return res


class HostFailed(CompmakeException):
    """ The job has been interrupted and must
        be redone (it has not failed, though) """

    def __init__(self, host, job_id, reason, bt):
        self.host = host
        self.job_id = job_id
        self.reason = reason
        self.bt = bt
        self.msg = "Host %r failed for %r: %s\n%s" % (
            self.host,
            self.job_id,
            self.reason,
            indent(self.bt, "|"),
        )
        CompmakeException.__init__(self, self.msg, host=host, reason=reason, bt=bt)

    def __str__(self):
        return self.msg

    def get_result_dict(self):
        res = dict(
            abort=f"Host failed for {self.job_id!r}.",
            host=self.host,
            job_id=self.job_id,
            reason=self.reason,
            bt=self.bt,
        )
        return res

    @staticmethod
    def from_dict(res):
        from compmake.jobs.result_dict import result_dict_check

        result_dict_check(res)
        try:
            _ = res["abort"]
            e = HostFailed(host=res["host"], job_id=res["job_id"], bt=res["bt"], reason=res["reason"])
        except KeyError as e:
            raise_wrapped(CompmakeBug, e, "Incomplete dict", res=res, keys=list(res.keys()))

        return e
